<?php

namespace Drupal\freecaster\Fcapi;

/**
 * Custom Exception for freecaster.
 */
class FCException extends \Exception
{
}


/**
 * Official Freecaster API.
 *
 * @see https://doc.fcst.tv/Platform/API%20Reference/
 * @author Yannick Delwiche
 * @copyright (c) 2012-2020 Freecaster SPRL
 *
 * @package Freecaster Engine
 *
 * @method create_channel(int $parent_id = NULL, array $params = [])
 * @method create_video(int $parent_id = NULL, array $params = [])
 * @method create_chapter(int $video_id, array $params)
 * @method delete_channel(int $channel_id)
 * @method delete_video(int $video_id)
 * @method delete_chapter(int $chapter_id)
 * @method get_channel(int $channel_id)
 * @method get_channels(array $filters)
 * @method get_hierarchy(int $channel_id = NULL)
 * @method get_streaming_config(int $video_id)
 * @method get_video(int $video_id)
 * @method get_videos(array $filters)
 * @method get_chapters(int $video_id)
 * @method link_channel_to_channel(int $child_id, int $parent_id)
 * @method link_video_to_channel(int $video_id, int $channel_id)
 * @method set_posterframe(int $video_id, array $params)
 * @method set_streams(int $video_id, array $params)
 * @method unlink_channel_from_channel(int $child_id, int $parent_id)
 * @method unlink_video_from_channel(int $video_id, int $channel_id)
 * @method update_channel(int $channel_id, array $params)
 * @method update_video(int $video_id, array $params)
 * @method update_chapter(int $chapter_id, array $params)
 * @method upload_video(int $video_id, array $params)
 */
class FCAPI {

  public $num_records = NULL;

  public $headers_out = NULL;

  public $body_out = NULL;

  public $headers_in = NULL;

  public $body_in = NULL;

  private $entrypoint = 'https://freecaster.tv/api/';

  private $user_id;

  private $api_key;

  private $format;

  private $debug;

  public function __construct(
    $user_id = 0,
    $api_key = '',
    $entrypoint = NULL,
    $debug = FALSE,
    $format = 'json'
  ) {
    $this->user_id = $user_id;
    $this->api_key = $api_key;

    $this->format = ($format == 'xml') ? 'text/xml' : 'application/json';
    $this->debug = $debug;

    if (!empty($entrypoint)) {
      $this->entrypoint = $entrypoint;
    }
  }

  /**
   * @param string $method
   * @param array $data
   *
   * @return mixed
   *
   * @throws FCException
   */
  public function __call($method, $data = []) {
    $url = $method;

    if (count($data) >= 1) {
      while ((count($data) > 0) && (!is_array($data[0]))) {
        $url .= '/' . array_shift($data);
      }
      if (count($data) > 0) {
        if (is_array($data[0])) {
          $data = $data[0];
        }
        if (isset($data[0])) {
          $data = ['data' => $data];
        }
      }
    }

    // Add the user ID parameter.
    $data['fc_uid'] = $this->user_id;
    // Add the timestamp.
    $data['ts'] = time();

    // Compute the normalized call.
    ksort($data);
    $normalized_data = [];
    foreach ($data as $key => &$val) {
      // Empty values should be explicitly
      // set to not be filtered out by the cURL call.
      if ($val === NULL) {
        $val = 'null';
      }
      // Do not include file uploads in the normalized call verification.
      if (($val) && (is_string($val)) && ($val[0] == '@')) {
        $file = substr($val, 1);
        if (empty($file)) {
          unset($data[$key]);
          continue;
        }
        $val = curl_file_create($file);
        continue;
      }
      if ((is_array($val)) || (is_object($val))) {
        $val = json_encode($val);
      }
      $normalized_data[$key] = $val;
    }
    $normalized_call = $url . '?' . http_build_query($normalized_data);

    // Set the call signature.
    $data['fc_sig'] = sha1($this->api_key . $normalized_call);

    // Encode characters and restore path separator.
    $url = rawurlencode($url);
    $url = str_replace('%2F', '/', $url);

    // If some values are array, we need the URL-encoded string.
    foreach ($data as &$val) {
      if (is_array($val)) {
        $data = http_build_query($data);
        break;
      }
    }

    $this->num_records = NULL;

    // Call the API server.
    $curl = curl_init($this->entrypoint . $url);
    curl_setopt_array(
      $curl,
      [
        CURLOPT_POST           => TRUE,
        CURLOPT_HTTP_VERSION   => CURL_HTTP_VERSION_1_0,
        CURLOPT_RETURNTRANSFER => TRUE,
        CURLOPT_HTTPHEADER     => ['Accept: ' . $this->format],
        CURLOPT_POSTFIELDS     => $data,
        CURLOPT_CONNECTTIMEOUT => 120,
        CURLOPT_TIMEOUT        => 600,
        CURLOPT_SSL_VERIFYPEER => FALSE,
      ]
    );
    if ($http_proxy = $this->get_proxy()) {
      curl_setopt($curl, CURLOPT_PROXY, $http_proxy);
    }
    if ($this->debug) {
      curl_setopt_array(
        $curl,
        [
          CURLOPT_HEADER      => TRUE,
          CURLINFO_HEADER_OUT => TRUE,
        ]
      );
      if (empty($_COOKIE['XDEBUG_SESSION'])) {
        curl_setopt($curl, CURLOPT_COOKIE, 'XDEBUG_SESSION=DEBUG');
      }
    }
    else {
      curl_setopt($curl, CURLOPT_HEADER, FALSE);
    }
    $result = curl_exec($curl);
    $errno = curl_errno($curl);
    $error = curl_error($curl);
    if ($this->debug) {
      $this->headers_out = curl_getinfo($curl, CURLINFO_HEADER_OUT);
      $this->body_out    = (is_array($data)) ? http_build_query($data) : $data;
    }
    curl_close($curl);

    if (!empty($error)) {
//      throw new FCException('Error ' . $errno . ': ' . $error);
    }

    if ($this->debug) {
      [$this->headers_in, $this->body_in] = explode("\r\n\r\n", $result);
      $result = $this->body_in;
    }

    // Decode the result.
    $result = @json_decode($result);
    // Check for errors.
    if (empty($result)) {
//      throw new FCException('Unknown API Error');
    }
    if (!empty($result->error)) {
//      throw new FCException('API Error: ' . $result->error);
    }
    // Return true if we received a simple "OK".
    if ((!empty($result->info)) && ($result->info == 'OK')) {
      return TRUE;
    }

    // If we requested objects, return them instead of the server reply.
    [$action, $request] = preg_split('#[_/]#', $method, 3);
    if ((($action == 'get') || ($action == 'create'))
      && (isset($result->$request))
    ) {
      if (isset($result->num_records)) {
        $this->num_records = $result->num_records;
      }
      return $result->$request;
    }

    return $result;
  }

  /**
   * Setter for all properties.
   *
   * @param string $name
   *   The name of the property.
   * @param string $value
   *   The value implemented to the property.
   */
  public function __set($name, $value) {
    $this->$name = $value;
  }

  /**
   * Get the web server proxy settings.
   *
   * @return string
   *   Proxy host:port
   */
  private function get_proxy() {
    if (!empty($_ENV['http_proxy'])) {
      return $_ENV['http_proxy'];
    }
    else {
      if (!empty($_SERVER['http_proxy'])) {
        return $_SERVER['http_proxy'];
      }
    }
    return '';
  }

}
