<?php

namespace Drupal\freecaster\Form;

use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\State\StateInterface;
use Drupal\freecaster\Services\FreecasterService;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class UserInfoForm.
 *
 * @package Drupal\freecaster\Form
 */
final class UserInfoForm extends ConfigFormBase {

  const FC_CACHE_TTL = 600;
  const FC_API_ENTRYPOINT = 'https://freecaster.tv/api/';

  /**
   * Date Formatter service.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected $dateFormatter;

  /**
   * State service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * Freecaster service.
   *
   * @var \Drupal\freecaster\Services\FreecasterService
   */
  protected $freecasterService;

  /**
   * {@inheritdoc}
   */
  public function __construct(ConfigFactoryInterface $config_factory, DateFormatterInterface $date_formatter, StateInterface $state, FreecasterService $freecaster_service) {
    parent::__construct($config_factory);
    $this->dateFormatter = $date_formatter;
    $this->state = $state;
    $this->freecasterService = $freecaster_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('date.formatter'),
      $container->get('state'),
      $container->get('freecaster.api')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'fc_user_info_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $config = $this->config('freecaster.settings');

    $form['fc_user_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('API user ID'),
      '#description' => $this->t('Freecaster User ID'),
      '#maxlength' => 64,
      '#size' => 64,
      '#default_value' => $config->get('fc_user_id'),
      '#required' => TRUE,
    ];
    $form['fc_user_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('API user Key'),
      '#description' => $this->t('Freecaster User Key'),
      '#maxlength' => 255,
      '#size' => 64,
      '#default_value' => $config->get('fc_user_key'),
      '#required' => TRUE,
    ];

    $form['fc_user_entrypoint'] = [
      '#type'          => 'textfield',
      '#title'         => $this->t("Freecaster API entrypoint"),
      '#default_value' => $config->get('fc_user_entrypoint') ?: self::FC_API_ENTRYPOINT,
      '#size'          => 64,
      '#description'   => $this->t("The URL used to contact the Freecaster Platform API."),
      '#required'      => TRUE,
    ];

    $form['fc_cache_ttl'] = [
      '#type'          => 'textfield',
      '#title'         => t("Cache lifetime"),
      '#default_value' => $config->get('fc_cache_ttl') ?: self::FC_CACHE_TTL,
      '#size'          => 2,
      '#description'   => $this->t("The number of seconds to keep Freecaster Platform data in the cache."),
      '#required'      => FALSE,
    ];

    $last_update = $this->state->get('fc_last_connection');
    $form['fc_last_connection'] = [
      '#type' => 'item',
      '#title' => $this->t('Last connection'),
      '#markup' => $last_update ? $this->dateFormatter->format($last_update) : $this->t('Never'),
    ];

    if ($account_channel = $config->get('fc_channel_account')) {
      $form['fc_channel_account'] = [
        '#type' => 'item',
        '#title' => $this->t('Account channel'),
        '#markup' => $this->formatChannelConfig($account_channel),
      ];
    }

    $form['actions']['check_api'] = [
      '#type' => 'submit',
      '#value' => $this->t('Check API credentials'),
      "#weight" => 2,
      '#button_type' => 'secondary',
      '#submit' => ['::checkCredentials'],
    ];

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $fc_entry_point = $form_state->getValue('fc_user_entrypoint');

    // Check API entry point URL format.
    if (!UrlHelper::isValid($fc_entry_point, TRUE)) {
      $form_state->setErrorByName('fc_user_entrypoint', $this->t('Please use a valid URL'));
    }

    parent::validateForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $config = $this->config('freecaster.settings');
    $config
      ->set('fc_user_id', $form_state->getValue('fc_user_id'))
      ->set('fc_user_key', $form_state->getValue('fc_user_key'))
      ->set('fc_user_entrypoint', $form_state->getValue('fc_user_entrypoint'))
      ->set('fc_cache_ttl', $form_state->getValue('fc_cache_ttl'))
      ->save();

    $this->checkCredentials($form, $form_state);
    parent::submitForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['freecaster.settings'];
  }

  /**
   * Check if credentials are valid.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return bool
   *   Return True if the credentials are valid.
   */
  public function checkCredentials(array &$form, FormStateInterface $form_state) {
    $form_state->setRebuild();

    $fc_user_id = $form_state->getValue('fc_user_id');
    $fc_user_key = $form_state->getValue('fc_user_key');
    $fc_user_entrypoint = $form_state->getValue('fc_user_entrypoint');
    $fc_error = "";

    if (!\Drupal::service('freecaster.api')->isConfigValid($fc_user_id, $fc_user_key, $fc_user_entrypoint, $fc_error)) {
      $this->messenger()->addError($fc_error);
      return FALSE;
    }
    else {
      $this->messenger()->addMessage($this->t('The FreeCaster API credentials are valid'));
      return TRUE;
    }
  }

  /**
   * Format channel config array to display it as a string.
   *
   * @param array $base_channel
   *   The base channel array (keys: id;name;).
   *
   * @return string
   *   Return a well formated string.
   */
  private function formatChannelConfig(array $base_channel) {
    return implode(': ', $base_channel);
  }

  /**
   * Format the select option for the base channel select.
   *
   * @param array $hierarchy
   *   The channel hierarchy array provided by Freecaster.
   * @param int $level
   *   The channel depth level.
   *
   * @return array
   *   Return an array of options.
   */
  private function formatChannelHierarchOption(array $hierarchy, int $level = 1) {
    $options = [];
    foreach ($hierarchy as $key => $details) {
      $options[$key] = str_repeat('-', $level) . $details['name'];
      if (!empty($details['children'])) {
        $options = $options + $this->formatChannelHierarchOption($details['children'], $level + 1);
      }
    }
    return $options;
  }

}
