<?php

namespace Drupal\freecaster\Plugin\Field\FieldFormatter;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\FormatterBase;
use Drupal\Core\Form\FormStateInterface;

/**
 * Plugin implementation of the Freecaster Player formatter.
 *
 * @FieldFormatter(
 *   id = "fcplayer",
 *   label = "Freecaster Player",
 *   field_types = {
 *     "integer",
 *   },
 * )
 */
class FreecasterPlayerFormatter extends FormatterBase {

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $element = [];

    foreach ($items as $delta => $item) {

      $video_id = $item->value;

      // Display placeholder div tag.
      $element[$delta]['placeholder'] = [
        '#type'       => 'html_tag',
        '#tag'        => 'div',
        '#attributes' => [
          'id' => $this->getContainerId($video_id),
        ],
      ];

      // Display related script.
      $element[$delta]['script'] = [
        '#type'       => 'html_tag',
        '#tag'        => 'script',
        '#attributes' => [
          'src' => $this->getPlayerUrl($video_id),
        ],
      ];

    }

    return $element;
  }

  /**
   * Get the url including options for the Freecaster player script.
   *
   * @param int $vid
   *   The ID of the video.
   *
   * @return string
   *   Return the full script player URL.
   */
  protected function getPlayerUrl(int $vid) {
    $base_url = new FormattableMarkup('//player.freecaster.com/embed/@video_id.js', ['@video_id' => $vid]);

    // Add configuration to player.
    $configs = $this->getSettings();

    $args = ['id' => $this->getTagId($vid)];
    foreach ($configs as $key => $value) {
      if (is_bool($value)) {
        $value = intval($value);
      }
      $args[$key] = $value;
    }

    return $base_url . '?' . UrlHelper::buildQuery($args);

  }

  /**
   * Get ID for The player configuration and container prefix.
   *
   * @param int $vid
   *   The ID of the video.
   *
   * @return string
   *   The ID to use for the tag replacement.
   */
  protected function getTagId(int $vid) {
    return 'video' . $vid;
  }

  /**
   * Get ID for The player div container.
   *
   * @param int $vid
   *   The ID of the video.
   *
   * @return string
   *   The ID to use for the html tag.
   */
  protected function getContainerId(int $vid) {
    return $this->getTagId($vid) . '_container';
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(FieldDefinitionInterface $field_definition) {
    if ($field_definition->getTargetEntityTypeId() !== 'media') {
      return FALSE;
    }

    if (parent::isApplicable($field_definition)) {
      return $field_definition->getTargetBundle() == 'freecaster_video';
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'width' => 0,
      'height' => 0,
      'muted' => FALSE,
      'autoplay' => FALSE,
      'loop' => FALSE,
      'controls' => TRUE,
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    return parent::settingsForm($form, $form_state) + [
      'width' => [
        '#type' => 'number',
        '#title' => $this->t('The desired width of your video player in pixels'),
        '#default_value' => $this->getSetting('width'),
        '#size' => 5,
        '#maxlength' => 5,
        '#field_suffix' => $this->t('pixels'),
        '#min' => 0,
      ],
      'height' => [
        '#type' => 'number',
        '#title' => $this->t('The desired height of your video player in pixels'),
        '#default_value' => $this->getSetting('height'),
        '#size' => 5,
        '#maxlength' => 5,
        '#field_suffix' => $this->t('pixels'),
        '#min' => 0,
      ],
      'muted' => [
        '#type' => 'checkbox',
        '#title' => $this->t('The player should be muted during playback'),
        '#default_value' => $this->getSetting('muted'),
      ],
      'autoplay' => [
        '#type' => 'checkbox',
        '#title' => $this->t('The player will attempt to begin playback automatically when a page is loaded'),
        '#default_value' => $this->getSetting('autoplay'),
      ],
      'loop' => [
        '#type' => 'checkbox',
        '#title' => $this->t('The player should loop content after a playlist completes'),
        '#default_value' => $this->getSetting('loop'),
      ],
      'controls' => [
        '#type' => 'checkbox',
        '#title' => $this->t('Display the video controls (controlbar, display icons and dock buttons)'),
        '#default_value' => $this->getSetting('controls'),
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = parent::settingsSummary();

    // Summary for fixed size.
    if ($this->getSetting('width') && $this->getSetting('height')) {
      $summary[] = $this->t('%width x %height pixels', [
        '%width' => $this->getSetting('width'),
        '%height' => $this->getSetting('height'),
      ]);
    }
    elseif ($this->getSetting('width')) {
      $summary[] = $this->t('Width: %width pixels', [
        '%width' => $this->getSetting('width'),
      ]);
    }
    elseif ($this->getSetting('height')) {
      $summary[] = $this->t('Height: %height pixels', [
        '%height' => $this->getSetting('height'),
      ]);
    }

    $summary[] = $this->t('Muted: %choice', [
      '%choice' => $this->getSetting('muted') ? $this->t('Yes') : $this->t('No'),
    ]);

    $summary[] = $this->t('Autoplay: %choice', [
      '%choice' => $this->getSetting('autoplay') ? $this->t('Yes') : $this->t('No'),
    ]);

    $summary[] = $this->t('Repeat: %choice', [
      '%choice' => $this->getSetting('loop') ? $this->t('Yes') : $this->t('No'),
    ]);

    $summary[] = $this->t('Display controls bar: %choice', [
      '%choice' => $this->getSetting('controls') ? $this->t('Yes') : $this->t('No'),
    ]);

    return $summary;
  }

}
