<?php

namespace Drupal\freecaster\Plugin\Field\FieldWidget;

use Drupal\Component\Utility\Bytes;
use Drupal\Component\Utility\Environment;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldWidget\NumberWidget;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;

/**
 * Widget dedicated to add or upload a video from Freecaster.
 *
 * @FieldWidget(
 *   id = "freecaster_id",
 *   label = @Translation("Freecaster video"),
 *   field_types = {
 *      "integer"
 *   },
 * )
 */
class FreecasterUploadWidget extends NumberWidget {

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'placeholder' => '',
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(
    FieldItemListInterface $items,
    $delta,
    array $element,
    array &$form,
    FormStateInterface $form_state
  ) {
    // Do not allow to upload video on the edit form.
    if ($value = isset($items[$delta]->value) ? $items[$delta]->value : NULL) {
      return parent::formElement($items, $delta, $element, $form, $form_state);
    }

    $elements = [
      '#type' => 'fieldgroup',
      '#title' => 'Sources',
      '#attributes' => [
        'class' => 'fc-video-id--wrapper',
      ],
      '#element_validate' => [
        [$this, 'validate'],
      ],
    ];

    // Checkbox to select ID or VIDEO.
    $elements['upload_method'] = [
      '#title' => 'Choose your method',
      '#type' => 'radios',
      '#options' => [$this->t('Upload a video'), $this->t('Enter the Freecaster video ID')],
      '#default_value' => 0,
    ];

    // Upload video.
    $max_filesize = Bytes::toInt(Environment::getUploadMaxSize());
    $elements['file_container'] = [
      '#type' => 'container',
      '#states' => [
        'invisible' => [
          ':input[name="field_video_fcid[0][value][upload_method]"]' => ['value' => 1],
        ],
      ],
    ];

    $elements['file_container']['file'] = [
      '#title' => 'Upload your video',
      '#type' => 'managed_file',
      '#upload_location' => 'temporary://freecaster_video/',
      '#upload_validators' => [
        'file_validate_extensions' => ['mp4 avi ogv'],
        'file_validate_size' => [$max_filesize],
      ],
      '#file_exists_behavior' => FileSystemInterface::EXISTS_REPLACE,
    ];

    $elements['file_container']['file_upload_description'] = [
      '#theme' => 'file_upload_help',
      '#description'  => '',
      '#upload_validators' => $elements['file_container']['file']['#upload_validators'],
    ];

    // Video ID.
    unset($element['#weight']);
    unset($element['#required']);

    /** @var \Drupal\media\MediaForm $media_form */
    $media_form = $form_state->getformObject();
    // Hide the video ID field for the add form.
    $process = $media_form->getOperation() == 'add' ? 'visible' : 'enabled';

    $elements['id'] = $element + [
      '#title' => 'Freecaster video ID',
      '#type' => 'number',
      '#default_value' => $value,
      '#states' => [
        $process => [
          ':input[name="field_video_fcid[0][value][upload_method]"]' => ['value' => 1],
        ],
      ],
    ];

    return ['value' => $elements];
  }

  /**
   * Validate the dedicated Freecaster video ID field.
   *
   * @param array $element
   *   The form element to validate.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   */
  public function validate(array $element, FormStateInterface $form_state) {
    // Validation is depending of the chosen method.
    $video_id = $element['id']['#value'];

    if ($element['upload_method']['#value']) {
      if (empty($element['id']['#value'])) {
        $form_state->setError($element['id'], $this->t("Please enter the video ID"));
      }
    }
    else {
      if (empty($element['file_container']['file']['#value']['fids'])) {
        $form_state->setError($element['file_container']['file'], $this->t("Please upload a video"));
      }
      else {
        $form_state->setTemporaryValue('video', reset($element['file_container']['file']['#files']));
        $video_id = -1;
      }
    }

    // Set the value of the entire form element.
    if ($video_id) {
      $form_state->setValueForElement($element, $video_id);
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(FieldDefinitionInterface $field_definition) {
    if ($field_definition->getTargetEntityTypeId() !== 'media') {
      return FALSE;
    }

    if (parent::isApplicable($field_definition)) {
      return $field_definition->getTargetBundle() == 'freecaster_video';
    }

    return FALSE;
  }

}
