<?php

namespace Drupal\freecaster\Plugin\media\Source;

use Drupal\Component\Utility\Crypt;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldTypePluginManagerInterface;
use Drupal\Core\File\Exception\FileException;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\freecaster\Services\FreecasterService;
use Drupal\media\MediaInterface;
use Drupal\media\MediaSourceBase;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Freecaster video entity media source.
 *
 * @see \Drupal\file\FileInterface
 *
 * @MediaSource(
 *   id = "freecaster_video",
 *   label = @Translation("Freecaster video"),
 *   description = @Translation("Use remote video hosted on Freecaster."),
 *   allowed_field_types = {"integer"},
 *   thumbnail_alt_metadata_attribute = "alt",
 *   default_thumbnail_filename = "freecaster.png"
 * )
 */
class FreecasterVideo extends MediaSourceBase {

  /**
   * The logger channel for media.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The file system.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The Freecaster API service.
   *
   * @var \Drupal\freecaster\Services\FreecasterService
   */
  protected $freecasterService;

  /**
   * The http client service.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * Construct Freecaster video Source.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\freecaster\Services\FreecasterService $freecaster_service
   *   The Freecaster service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The config factory service.
   * @param \Drupal\Core\Field\FieldTypePluginManagerInterface $field_type_manager
   *   The entity field manager service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The field type plugin manager service.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger channel for media.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The HTTP client.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    $plugin_definition,
    FreecasterService $freecaster_service,
    EntityTypeManagerInterface $entity_type_manager,
    EntityFieldManagerInterface $entity_field_manager,
    FieldTypePluginManagerInterface $field_type_manager,
    ConfigFactoryInterface $config_factory,
    LoggerInterface $logger,
    MessengerInterface $messenger,
    ClientInterface $http_client,
    FileSystemInterface $file_system = NULL
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $entity_type_manager,
      $entity_field_manager,
      $field_type_manager,
      $config_factory
    );
    $this->freecasterService = $freecaster_service;
    $this->logger = $logger;
    $this->messenger = $messenger;
    $this->httpClient = $http_client;
    $this->fileSystem = $file_system;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('freecaster.api'),
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('plugin.manager.field.field_type'),
      $container->get('config.factory'),
      $container->get('logger.factory')->get('media'),
      $container->get('messenger'),
      $container->get('http_client'),
      $container->get('file_system')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getMetadataAttributes() {
    return [
      'title' => $this->t('Resource title'),
      'channel_id' => $this->t('The channel id from the provider'),
      'default_name' => $this->t('Default name of the media item'),
      'status' => $this->t('The status of the resource'),
      'url' => $this->t('The source URL of the resource'),
      'duration' => $this->t('The duration of the resource'),
      'link' => $this->t('The link to the video'),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getMetadata(MediaInterface $media, $attribute_name) {

    // Get the video ID from the source field.
    $video_id = $this->getSourceFieldValue($media);

    if (!$video_id) {
      return FALSE;
    }

    try {
      $remote_video = $this->freecasterService->getVideo($video_id);
    }
    catch (\Exception $exception) {
      $this->logger->error($exception->getMessage());
    }

    switch ($attribute_name) {
      case 'default_name':
        if ($title = $this->getMetadata($media, 'title')) {
          return $title;
        }
        elseif ($url = $this->getMetadata($media, 'url')) {
          return $url;
        }
        return parent::getMetadata($media, 'default_name');

      case 'thumbnail_uri':
        if ($remote_video && !isset($remote_video->thumbnail)) {
          // Processing in progress.
          if ($remote_video->status == -1 && $remote_video->id) {
            // Queue the thumbnail creation.
            $data = [
              'created' => $media->get('created')->value,
              'video_id' => $remote_video->id,
              'remote_video' => $remote_video,
            ];
            $this->freecasterService->setQueueItem($data);
          }
          return NULL;
        }
        return $this->getLocalThumbnailUri($remote_video->thumbnail);

      case 'title':
        return $remote_video->name;

      case 'channel_id':
        return $remote_video->id;

      case 'status':
        return $remote_video->status;

      case 'url':
        return $remote_video->url;

      case 'duration':
        return $remote_video->duration;

      case 'link':
        return $remote_video->link;

      default:
        break;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'thumbnails_directory' => 'public://freecaster_thumbnails',
    ] + parent::defaultConfiguration();
  }

  /**
   * Returns the local URI for a resource thumbnail.
   *
   * Inspired by the oembed behaviour.
   *
   * @param string $remote_thumbnail_url
   *   The external thumbnail uri provided by the Freecaster API.
   *
   * @return string|null
   *   The local thumbnail URI, or NULL if it could not be downloaded, or if the
   *   resource has no thumbnail at all.
   */
  protected function getLocalThumbnailUri(string $remote_thumbnail_url) {

    // Remove the query string, since we do not want to include it in the local
    // thumbnail URI.
    $local_thumbnail_url = parse_url($remote_thumbnail_url, PHP_URL_PATH);

    // Compute the local thumbnail URI, regardless of whether or not it exists.
    $configuration = $this->getConfiguration();
    $directory = $configuration['thumbnails_directory'];
    $local_thumbnail_uri = "$directory/" . Crypt::hashBase64($local_thumbnail_url) . '.' . pathinfo($local_thumbnail_url, PATHINFO_EXTENSION);

    // If the local thumbnail already exists, return its URI.
    if (file_exists($local_thumbnail_uri)) {
      return $local_thumbnail_uri;
    }

    // The local thumbnail doesn't exist yet, so try to download it. First,
    // ensure that the destination directory is writable, and if it's not,
    // log an error and bail out.
    if (!$this->fileSystem->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS)) {
      $this->logger->warning('Could not prepare thumbnail destination directory @dir for freecaster media.', [
        '@dir' => $directory,
      ]);
      return NULL;
    }

    try {
      $response = $this->httpClient->get($remote_thumbnail_url);
      if ($response->getStatusCode() === 200) {
        $this->fileSystem->saveData((string) $response->getBody(), $local_thumbnail_uri, FileSystemInterface::EXISTS_REPLACE);
        return $local_thumbnail_uri;
      }
    }
    catch (RequestException $e) {
      $this->logger->warning($e->getMessage());
    }
    catch (FileException $e) {
      $this->logger->warning('Could not download remote thumbnail from {url}.', [
        'url' => $remote_thumbnail_url,
      ]);
    }
    return NULL;
  }

}
