<?php

namespace Drupal\freecaster\Services;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\EntityReferenceFieldItemList;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\State\StateInterface;
use Drupal\freecaster\Fcapi\FCAPI;
use Drupal\taxonomy\Entity\Term;
use Drupal\taxonomy\TermInterface;

/**
 * Service to manage Freecaster API calls and caches.
 */
class FreecasterService {

  /**
   * The Freecaster API.
   *
   * @var \Drupal\freecaster\Fcapi\FCAPI
   */
  private $fcApi;

  /**
   * Config service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The Entity Type Manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * State service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  private $state;

  /**
   * Cache backend service.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  private $cache;

  /**
   * The Logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  private $logger;

  /**
   * The Queue service.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  private $queueFactory;

  /**
   * Construct the Freecaster service.
   */
  public function __construct(ConfigFactoryInterface $config, EntityTypeManagerInterface $entity_type_manager, StateInterface $state, CacheBackendInterface $cache, LoggerChannelFactoryInterface $logger, QueueFactory $queue_factory) {
    $this->configFactory = $config;
    $this->entityTypeManager = $entity_type_manager;
    $this->state = $state;
    $this->cache = $cache;
    $this->logger = $logger->get('freecaster');
    $this->queueFactory = $queue_factory;

    $this->fcApi = new FCAPI();
    $this->setConfigFromSettings();
  }

  /**
   * Get Freecaster API.
   *
   * @return \Drupal\freecaster\Fcapi\FCAPI
   *   Retrun Freecaster API object.
   */
  public function getApi() {
    return $this->fcApi;
  }

  /**
   * Get the Freecaster settings.
   *
   * @return \Drupal\Core\Config\Config
   *   Return config related to Freecaster.
   */
  protected function getConfig(bool $editable = FALSE) {
    return $editable ? $this->configFactory->getEditable('freecaster.settings') : $this->configFactory->get('freecaster.settings');
  }

  /**
   * Check if API crendentials exist.
   *
   * @return bool
   *   Return true if no API access has been defined.
   */
  public function isConfigEmpty() {
    $config = $this->getConfig();
    return (!$config->get('fc_user_id') && !$config->get('fc_user_key'));
  }

  /**
   * Check if credentials are valid.
   *
   * @param string|null $fcapi_uid
   *   The Freecaster user id.
   * @param string|null $fcapi_key
   *   The Freecaster key.
   * @param string|null $fcapi_entry_point
   *   The Freecaster entry point address.
   * @param string|null $fcapi_error
   *   The Freecaster api error message.
   *
   * @return bool
   *   Return true if the API respond using the credentials.
   */
  public function isConfigValid(string $fcapi_uid = NULL, string $fcapi_key = NULL, string $fcapi_entry_point = NULL, string &$fcapi_error = NULL) {
    // Check existing config.
    if (!$fcapi_uid && !$fcapi_key) {
      $test = $this->fcApi->get_hierarchy();
    }
    else {
      $api = clone $this->fcApi;

      $api->user_id = $fcapi_uid;
      $api->api_key = $fcapi_key;
      if ($fcapi_entry_point) {
        $api->entrypoint = $fcapi_entry_point;
      }

      $test = $api->get_hierarchy();
      unset($api);
    }

    if (property_exists($test, 'error')) {
      $fcapi_error = $test->error;
      return FALSE;
    }

    // Store last connection time and first channel.
    $this->state->set('fc_last_connection', time());

    $channel_id = array_keys((array) $test)[0];
    if ($channel_id) {
      $channel = [
        'id' => $channel_id,
        'name' => $test->$channel_id->name,
      ];

      // Save configs.
      $config = $this->getConfig(TRUE);
      $config->set('fc_channel_account', $channel);
      $config->set('fc_channel_hierarchy', json_decode(json_encode($test->$channel_id->children), TRUE));
      $config->save();
    }

    return TRUE;
  }

  /**
   * Set Credentials from the settings.
   */
  protected function setConfigFromSettings() {
    $config = $this->getConfig();

    if (!$this->isConfigEmpty()) {
      $this->fcApi->user_id = $config->get('fc_user_id');
      $this->fcApi->api_key = $config->get('fc_user_key');
      if ($fcapi_entry_point = $config->get('fc_user_entrypoint')) {
        $this->fcApi->entrypoint = $fcapi_entry_point;
      }

    }
  }

  /**
   * Check API credentials.
   */
  public function checkApiCredentials() {
    if ($this->isConfigEmpty()) {
      \Drupal::messenger()->addMessage(t("Configure your Freecaster API credentials first"), 'error');
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Get the Freecaster cache expiration.
   *
   * @return int
   *   Return the timestamp of the expiration.
   */
  protected function getCacheExpiration() {
    return time() + intval($this->getConfig()->get('fc_cache_ttl'));
  }

  /**
   * Get video info from cache or API.
   *
   * @param int $video_id
   *   The Freecaser video id.
   * @param bool $no_cache
   *   Force loading from the API.
   *
   * @return false|object
   *   Return the video information in JSON.
   *
   * @throws \Exception
   *   Throw an exeption when API return an error.
   */
  public function getVideo(int $video_id, bool $no_cache = FALSE) {
    if (!$this->checkApiCredentials()) {
      return FALSE;
    }

    $cid   = 'freecaster_video_' . $video_id;
    $cache = $this->cache->get($cid);

    // Call the API if not cached.
    if (!empty($cache->data) && !$no_cache) {
      $remote_video = $cache->data;
    }
    else {
      $remote_video = $this->fcApi->get_video($video_id);

      if (property_exists($remote_video, 'error')) {
        throw new \Exception($remote_video->error);
      }

      $this->cache->set($cid, $remote_video, $this->getCacheExpiration());
    }

    return $remote_video;
  }

  /**
   * Get thumbnail update queue.
   *
   * @return \Drupal\Core\Queue\QueueInterface
   *   Return the thumbnail update queue.
   */
  protected function getQueue() {
    return $this->queueFactory->get('fc_thumbnail_update');
  }

  /**
   * Convert data and add a new item to the thumbnail update queue.
   *
   * @param array $data
   *   Raw data under the form of array.
   *
   * @return bool|int
   *   Return the queue task id or FALSE if an exception is thrown.
   */
  public function setQueueItem(array $data) {
    // Prepare the item.
    $item = new \stdClass();
    $item->created = $data['created'];
    $item->video_id = $data['video_id'];

    try {
      return $this->getQueue()->createItem($item);
    }
    catch (\Exception $exception) {
      $this->logger->error(
        'An error occurs when create a item in queues (@error)',
        ['@error' => $exception->getMessage()]
      );
      return FALSE;
    }
  }

  /**
   * Get Taxonomy term storage for channels.
   *
   * @return mixed
   *   Return the storage.
   */
  protected function getChannelStorage() {
    return $this->entityTypeManager->getStorage('taxonomy_term');
  }

  /**
   * Sync channels.
   */
  public function syncChannels() {
    if (!$this->isConfigValid()) {
      return NULL;
    }

    $channels = $this->getConfig()->get('fc_channel_hierarchy');
    $taxonomies = $this->getChannelStorage()->loadByProperties(['name' => 'fc_video_channels']);

    if ($channels) {
      $this->updateChannels($channels);
    }

  }

  /**
   * Recursive function to update channels tree.
   *
   * @param array $channels
   *   An array of channels under the form of tree.
   * @param array $parent
   *   Current chanels for children channels.
   */
  protected function updateChannels(array $channels, array $parent = []) {
    foreach ($channels as $channel_id => $channel) {
      $parents = [];
      if (!empty($parent)) {
        $parents[] = $parent['tid'];
      }

      $term = $this->getChannelTerm($channel_id);
      if (empty($term)) {
        // Create the taxonomy for channel.
        $term = Term::create(
          [
            'vid'    => 'fc_video_channels',
            'name'   => $channel['name'],
            'parent' => $parents,
            'field_channel_id' => $channel_id,
          ]
        );
        $term->save();
      }
      else {
        /** @var \Drupal\taxonomy\TermInterface $term */
        $term = reset($term);

        if ($channel['name'] !== $term->label()) {
          $term->set('name', $channel['name']);
          $term->save();
        }

      }

      if (!empty($channel['children'])) {
        $parent = [
          'tid' => $term->id(),
          'channel_id' => $channel_id,
        ];
        $this->updateChannels($channel['children'], $parent);
      }
    }
  }

  /**
   * Get the corresponding term id from a FC channel id.
   *
   * @param int $channel_id
   *   The Freecaster channel id.
   *
   * @return mixed
   *   Return an array of taxanomies.
   */
  protected function getChannelTerm(int $channel_id) {
    return $this->getChannelStorage()->loadByProperties(['field_channel_id' => $channel_id]);
  }

  /**
   * Get parents ids from channel.
   *
   * @param int $channel_id
   *   Freecaster channel id.
   *
   * @return array
   *   Return an array of channel ids.
   */
  protected function getChannelParents(int $channel_id) {
    return $this->getChannelParent($this->getApi()->get_hierarchy(), $channel_id);
  }

  /**
   * Recursive helper to get parents ids from a Freecaster Hierarchy.
   *
   * @param \stdClass $hierarchy
   *   The channel hierarchy as an object.
   * @param int $channel_id
   *   The channel id to find.
   *
   * @return array
   *   Return an array of channel ids.
   */
  protected function getChannelParent(\stdClass $hierarchy, int $channel_id) {
    $parents = [];
    foreach ($hierarchy as $key => $channel) {
      $children = (array) $channel->children;
      if (!empty($children)) {
        if (in_array($channel_id, array_keys($children))) {
          $parents[] = $key;
        }
        $channel_parent = $this->getChannelParent(
          $channel->children,
          $channel_id
        );
        $parents = array_unique(array_merge($parents, $channel_parent));
      }
    }

    return $parents;
  }

  /**
   * Get Freecaster Channel id from a taxonomy.
   *
   * @param \Drupal\taxonomy\TermInterface $term
   *   Taxonomy term object.
   *
   * @return null|int
   *   Return the field channel id. Return FALSE if not defined.
   */
  public function getChannelId(TermInterface $term) {
    if ($term->hasField('field_channel_id') && !$term->get('field_channel_id')->isEmpty()) {
      return $term->field_channel_id->value;
    }

    return NULL;
  }

  /**
   * Return the default Channel.
   *
   * @return int|null
   *   Return a channel id or FALSE if there is no default config.
   */
  public function getDefaultChannel() {
    return $this->getConfig()->get('fc_channel_account')['id'] ?? FALSE;
  }

  /**
   * Return the term id of Default Channel.
   *
   * @return int|null
   *   Return the default channel taxonomy term id.
   */
  public function getDefaultChannelTermId() {
    if ($terms = $this->getChannelTerm($this->getDefaultChannel())) {
      if (count($terms)) {
        return key($terms);
      }
    }

    return NULL;
  }

  /**
   * Create Channel on Freecaster.
   *
   * @param \Drupal\taxonomy\TermInterface $term
   *   The taxonomy term source.
   *
   * @return false|int
   *   Return channel id or false if an error occurs.
   */
  public function createChannel(TermInterface $term) {
    $params = [
      'name' => $term->label(),
    ];
    // Prepare parents and get default channels.
    $parents = $term->parent;
    $parent_channel_id = $this->getDefaultChannel();

    if (!empty($parents) && $parents instanceof EntityReferenceFieldItemList) {
      $parents = $parents->referencedEntities();
      if (count($parents)) {
        /** @var \Drupal\taxonomy\TermInterface $parent */
        $parent = array_shift($parents);
        if (!$parent->get('field_channel_id')->isEmpty()) {
          $parent_channel_id = $parent->field_channel_id->value;
        }
      }
      else {
        // Set default Channel as term parent.
        if ($tid = $this->getDefaultChannelTermId()) {
          $term->set('parent', $tid);
        }
      }
    }

    // Create a new channel on Freecaster.
    $channel = $this->fcApi->create_channel($parent_channel_id, $params);
    if (empty($channel) || property_exists($channel, 'error')) {
      return FALSE;
    }

    // Link to other channels if there is multiple parents.
    foreach ($parents as $parent) {
      $parent_channel_id = $parent->field_channel_id->value;
      if ($parent_channel_id) {
        $this->fcApi->link_channel_to_channel($channel->id, $parent_channel_id);
      }
    }

    return $channel->id;
  }

  /**
   * Update channel properties on Freecaster.
   *
   * @param \Drupal\taxonomy\TermInterface $term
   *   Taxonomy term object which contains information.
   *
   * @return false|int
   *   Return the updated channel id or FALSE if nothing has been updated.
   */
  public function updateChannel(TermInterface $term) {
    if (!$this->isConfigValid()) {
      return FALSE;
    }

    $params = [
      'name' => $term->label(),
    ];

    if ($channel_id = $this->getChannelId($term)) {
      $channel = $this->getApi()->update_channel($channel_id, $params);
    }
    else {
      return FALSE;
    }

    if (property_exists($channel, 'error')) {
      return FALSE;
    }

    return $channel->id;
  }

  /**
   * Update channel parents on Freecaster.
   *
   * @param \Drupal\taxonomy\TermInterface $term
   *   The taxonomy term object which contains the information.
   */
  public function updateChannelParent(TermInterface $term) {
    $parents           = $term->parent;
    $parent_channel_id = $this->getDefaultChannel();

    if ($channel_id = $this->getChannelId($term)) {
      // Get Freecaster channels ids.
      $channel_parents = $this->getChannelParents($channel_id);

      $parents     = $parents->referencedEntities();
      $new_parents = [];
      if (count($parents)) {
        /** @var \Drupal\taxonomy\TermInterface $parent */
        foreach ($parents as $parent) {
          if ($parent_channel_id = $this->getChannelId($parent)) {
            $new_parents[] = $parent_channel_id;
          }
        }
      }

      $default_channel_id = $this->getDefaultChannel();
      if (empty($new_parents)) {
        $new_parents[] = $default_channel_id;
      }

      // Link new parents channels.
      foreach (array_diff($new_parents, $channel_parents) as $channel_to_add) {
        $this->getApi()->link_channel_to_channel($channel_id, $channel_to_add);
      }

      // Unlink old parents channels.
      foreach (array_diff($channel_parents, $new_parents) as $channel_to_del) {
        $this->getApi()->unlink_channel_from_channel($channel_id, $channel_to_del);
      }

    }
  }

  /**
   * Delete a channel on Freecaster.
   *
   * @param \Drupal\taxonomy\TermInterface $term
   *   Taxonomy term object which contains the information.
   */
  public function deleteChannel(TermInterface $term) {
    if (!$this->isConfigValid()) {
      return FALSE;
    }

    if ($channel_id = $this->getChannelId($term)) {
      $this->getApi()->delete_channel($channel_id);
    }
  }

}
